/* vi:ts=4:sw=4
 *
 * VIM - Vi IMitation 		Extension for Kanji Frontend Processor
 *
 * Code Contributions By:	Atsushi Nakamura	ann@mrit.mei.co.jp
 */

/*
 * Code to handle kanji FEP conversion.
 */


#include "vim.h"

#ifdef FEP
#include "globals.h"
#include "proto.h"
#include "param.h"
#include "kanji.h"

#ifdef ONEW
#define KF_KANAKAN	0x100
#endif

static char fep_buf[MAX_COLUMNS + 1] = "", *fbp = fep_buf;

#ifdef ONEW

static char msgbuf[MAX_COLUMNS + 1];
static char kanakan_buf[MAX_COLUMNS + 1] = "";
static char hira_buf[MAX_COLUMNS + 1] = "";
static char *henkan_mode = NULL;
static int  kanakan_len = 0;
static int  freq_save = FALSE;

#define FUNCKEY(c)		(c | 0xffffff00)
	int
ONEW_GETCHAR()
{
	u_char c;

	c = (u_char)vgetc();

	if (c == K_ZERO || c == Ctrl('\\'))		/* convert key */
		return p_oh ? NUL : FUNCKEY(K_ZERO);

	if (c >= 0x80 && c <= 0x9d)				/* function key */
		return FUNCKEY(c);

	return c;
}

	int
ONEW_PEEKCHAR(timeout)
	int *timeout;
{	int c;

	if( c = inchar(TRUE,FALSE) ){
	}else{
		/*if( WaitForChar(*timeout) )
			*timeout = 0;
		else*/
			*timeout = -1;
	}
	return c;
}

ONEW_DISP_ROMKANMODE(mode, help, imode)
	char *mode, *help;
{
	int rk, len;
	char *mchrs;
	static char modebuf[20];
	static int  oldimode = -1;

	if (!mode)
	{
		oldimode = -1;
		return;
	}

	henkan_mode = mode;

	rk = Onew_RK_imode();

	if (State == CMDLINE)
	{
		cmd_message(FALSE, mode, kanakan_len);
		if (oldimode != imode)
		{ 
			oldimode = imode;
		    ONEW_MESSAGE(rk, "%s %s", help, Onew_version());
		}
		if (imode & KF_KANAKAN)
			windgoto(Rows - 1, Columns - 1);
		else 
			cursorcmd();
		return;
	}

	if (p_fm) sprintf(modebuf,"%s", mode);

	switch(State)
	{
	case INSERT:
		mchrs = "[]I";
		break;
	case REPLACE:
		mchrs = "<>R";
		break;
	case NOMAPPING:
		mchrs = "()-";
		break;
	default:
		mchrs = "//!";
		break;
	}

	if (p_fm)
	{
		*strchr( modebuf, '[') = *mchrs ++;
		*strrchr(modebuf, ']') = *mchrs ++;
		screen_msg(rk, modebuf);
	}

	if (oldimode != imode)
	{
		oldimode = imode;
		if (p_fm)
			sprintf(msgbuf, Recording ? "<%s>%s": "[%s]%s",
													help, Onew_version());
		else
			sprintf(msgbuf, Recording ? "%s<%c>%s %s": "%s%c%s %s",
									mode, *(mchrs + 2), help, Onew_version());

	    { int i;
		  for(i = 0; i < Columns - (p_ru ? 24 : 1); i++)
		    if (IsKanji(msgbuf[i])) i++;
		  msgbuf[i] = NUL;
		}

		if (rk && !p_fm)
		{
			outstr(T_TI);
			smsg(msgbuf);
			outstr(T_TP);
		}
		else
			smsg(msgbuf);
	}

	cursupdate();
	setcursor();

	if (imode & KF_KANAKAN)
		windgoto(Rows - 1, Columns - 1);
}

	int
ONEW_DISP_KANAHALVES(s)
	char *s;
{
	if (State == CMDLINE)
	{
		cmd_insertstr(s);
		cmd_delchars(strlen(s));
		cursorcmd();
	}
	else
	{
		int mode, n;
		char *cp;

		mode = State;
		State = INSERT;

		for(cp = s, n = 0; *cp; cp++, n++)
			insertchar(*cp, NUL);
		Curpos.col -= n;

		cursupdate();
		setcursor();
		updateline();

		for(cp = s; *cp; cp++)
			delchar(FALSE);

		State = mode;
	}
}

	int
ONEW_DISP_KANAKANB(so, left, cur, right)
	int so;
	char *left, *cur, *right;
{	int d;
	char *cp;

	int mode = State;

	sprintf(kanakan_buf, "%s%s%s", left, cur, right);

	if (State == CMDLINE)
	{
		Kconvlnum	  = -1;
		KconvStart = KconvEnd = 0;
		KconvAltStart = strlen(left);					/* Reverse start    */
		KconvAltEnd   = KconvAltStart + strlen(cur);	/* Reverse end      */
	}
	else
	{
		State = INSERT;

		Kconvlnum     = Curpos.lnum;
		KconvStart    = Curpos.col;						/* Underscore start */
		KconvAltStart = KconvStart    + strlen(left);	/* Reverse start    */
		KconvAltEnd   = KconvAltStart + strlen(cur);	/* Reverse end      */
		KconvEnd      = KconvAltEnd   + strlen(right);	/* Underscore end   */
	}

	kanakan_len = fep_insstr(kanakan_buf);
	fep_delchars(kanakan_len);

	State = mode;
	return 0;
}

int	onew_henkan = FALSE;	/* to avoid textwidth folding */

	char
ONEW_KANAKAN(ch)
	char ch;
{
	char com, *cp;

	if (fbp != fep_buf)
	{
		fep_delchars(strlen(fep_buf));
		strncpy(hira_buf, fep_buf, fbp - fep_buf);
		hira_buf[fbp - fep_buf] = NUL;

		com = Onew_kanakan(ch, fep_buf, 0, fbp - fep_buf);
		freq_save = TRUE;

		strcpy(fep_buf, kanakan_buf);
		fbp = fep_buf + fep_insstr(fep_buf);

		ONEW_KAKUTEI(com);
		if (com != ESC && com != TAB) com = 0;
	}
	else
		com = ch;

	return com;
}

	int
ONEW_KAKUTEI(ch)
{
	int this_ch_was_used_for_kakutei = FALSE;

	if (fbp != fep_buf)
	{
		if (ch == BS || ch == DEL)
		{
			fep_delchars(fbp - fep_buf);

			strcpy(fep_buf, hira_buf);
			fbp = fep_buf + strlen(fep_buf);
	
			if (State == CMDLINE)
			{
				Kconvlnum     =  0;
				KconvAltStart = 
				KconvAltEnd   = -1;
			}
			else
			{
				Kconvlnum  = Curpos.lnum;
				KconvStart = Curpos.col;
				KconvEnd   = KconvStart + (fbp - fep_buf);
			}

			fep_insstr(fep_buf);
		}
		else
		{
			onew_henkan = FALSE;
			*fbp = NUL;
			AppendToRedobuff(fep_buf);
			onew_henkan = TRUE;

			*(fbp = fep_buf)  = hira_buf[0] = NUL;
			this_ch_was_used_for_kakutei = TRUE;
			Kconvlnum = 0;
		}
	}

	if (State == CMDLINE)
		redrawcmdline();
	else
	{
		if (p_fm && henkan_mode)
			screen_msg(TRUE, henkan_mode);
		cursupdate();
		setcursor();
		updateline();
	}


	kanakan_len = 0;
	return this_ch_was_used_for_kakutei;
}

ONEW_GOT_2BCHAR(buf, a, b)
	char *buf, a, b;
{
	sprintf(buf, "%c%c", a, b);
}

ONEW_MESSAGE_COLS()
{
	return p_ru ? Columns - 24 : Columns - 1;
}


ONEW_MESSAGE(so, form, a, b, c, d)
	char *form;
{
	sprintf(msgbuf, form, a, b, c, d);

    { int i;
	  for(i = 0; i < Columns - (p_ru ? 24 : 1); i++)
	    if (IsKanji(msgbuf[i])) i++;
	  msgbuf[i] = NUL;
	}

	if (State == CMDLINE)
		screen_msg(so, msgbuf);
	else
	{
		if (so) outstr(T_TI);
		smsg(msgbuf);
		if (so) outstr(T_TP);
	}
}

ONEW_BEEP()
{
	beep();
}

	int
fep_insstr(s)
	char *s;
{
	int len;
	int sm;

	sm = p_sm;				/* turn off  showmatch option */
	p_sm = FALSE;

	if (State == CMDLINE)
		len = cmd_insertstr(s);
	else
	{
		int w;

		for(len = 0; *s; len += w, s += w)
			w = insertchar(*s, *(s + 1));

		cursupdate();
		setcursor();
		updateline();
	}

	p_sm = sm; 				/* restore showmatch option */
	return len;
}

fep_delchars(len)
{
	if (State == CMDLINE)
	{
		cmd_delchars(len);
	}
	else
	{
		for(; len; len--)
		{
			Curpos.col--;
			delchar(FALSE);
		}
	}
}

#endif /* ONEW */

/*
 *	vim FEP interface
 */

fep_mode_switch(mode)
	int mode;
{
#ifdef ONEW
	{
		int onew_mode;

		Onew_RK_init();
		switch(mode)
		{
		case 0:
			onew_mode = '@';
			break;
		case -1:
			onew_mode = 'h';
			break;
		default:
			onew_mode = mode;
			break;
		}
		Onew_RK_cmode_set(onew_mode);
	}
#else
	return;
#endif
}

	int
fep_mode()
{
#ifdef ONEW
	{
		int mode;
		mode = Onew_RK_cmode();
		if (mode == '@') return 0;
		if (mode == 'h') return -1;
		return mode;
	}
#else
	return 0;
#endif
}

fep_freqsave()
{
#ifdef ONEW
	if (freq_save)
	{
		smsg("Saving frequency information.");
		Onew_KK_freqsave();
		smsg("");
		freq_save = FALSE;
	}
#endif
}

fep_getchar(k1, k2)
	char *k1, *k2;
{
#ifdef ONEW
	{
	    int oin;
		u_char in, ink;
		char *cp;
		static int first = TRUE;

		if (first)
		{
			*(fbp = fep_buf) = hira_buf[0] = NUL;
			first = FALSE;
		}

		if (onew_henkan)
			updateline();
		else
			onew_henkan = TRUE;

		ink = NUL;
		oin = Onew_romkan();

		if (oin == -1)
			mch_windexit(5);

		if (oin & 0xffffff00)
		{
			ONEW_KAKUTEI(in);
			Kconvlnum  = 0;
			*fbp = '\0';

			*k1 = oin;
			*k2 = '\0';
			return;
		}

		in = (char) oin;

		switch(in)
		{
		case Ctrl('C'):
			in = 0;
		case '\n': case '\r':
		case NUL:
			if (in == NUL) in = K_ZERO;
		case K_ZERO:
		case ESC:
		    ONEW_KAKUTEI(in);
			first = TRUE;
			ONEW_DISP_ROMKANMODE(NULL, "", -1);
			smsg("");
			onew_henkan = FALSE;
			if (State == CMDLINE)
				redrawcmdline();
			else
				if (insertchar(NUL, NUL))	/* format line */
					updateScreen(NOT_VALID);
				else
				{
					updateline();
					screen_msg(FALSE, NULL);
				}
			break;

		case Ctrl('U'):
			fep_delchars(fbp - fep_buf);
			*(fbp = fep_buf) = hira_buf[0] = NUL;
			if (State == CMDLINE)
				redrawcmdline();
			else
				updateline();
			break;

		case DEL:
		case BS:
			if (fbp != fep_buf)
			{
			    if (State == CMDLINE)
				{
					Kconvlnum  = 0;
					KconvStart = -1;
				}
				else
				{
					KconvStart = KconvEnd - (fbp - fep_buf);
					KconvEnd   = Curpos.col - 1;
			    }
				fbp--;
				if (IsKanji(*fbp))
				{
					fbp --;
					KconvEnd--;
				}
			}
			else
				Kconvlnum = 0;
			break;

		case ' ':
		case TAB:
			*fbp++ = in;
		case Ctrl('D'):
		case Ctrl('T'):
		case Ctrl('W'):
			ONEW_KAKUTEI(in);
			Kconvlnum  = 0;
			break;

		default:
			if (IsKanji(in))
			{
				ink = Onew_romkan();
				*fbp++ = in;
				*fbp++ = ink;
				KconvEnd   = Curpos.col + 1;
			}
			else if (' ' <= in && in < DEL)
			{
				*fbp++ = in;
				KconvEnd   = Curpos.col;
			}
			else
			{
				first = TRUE;
				if (in < ' ')
					ONEW_MESSAGE(TRUE,"Kanji/ONEW: '^%c' unknown  _o_",
																 in + '@');
				if (in >= DEL)
					ONEW_MESSAGE(TRUE,"Kanji/ONEW: x'%c' unkonwn _o_",
																 in & 0x7f);
				ONEW_BEEP(1);
				in = 0;
				break;
			}
			if (State == CMDLINE)
			{
				Kconvlnum =
				KconvAltStart = -1;
			}
			else
			{
				Kconvlnum  = Curpos.lnum;
				KconvStart = KconvEnd + 1 - (fbp - fep_buf);
				KconvAltStart = KconvAltEnd = Curpos.col;
			}
			break;
		}
		*fbp = '\0';

		*k1 = in;
		*k2 = ink;
		return;
	}

#else /* ONEW */
	{
		char *cp;
		if (!*fbp)
		{
		  	smsg("> ");
			settmode(0);
			for(cp = fep_buf; cp - fep_buf < MAX_COLUMNS; cp++)
			{
				*cp = vgetc();
				if (*cp == ESC || *cp == '\n' || *cp == '\r') break;
			}
/*			fgets(fep_buf, MAX_COLUMNS+1, stdin);
*/
			settmode(1);
			kanjiconvsfrom(fep_buf, KCODE_ANY);

			cp = strchr(fep_buf, '\n');
			if (cp) *cp = '\0';
			for(cp = fep_buf; *cp; cp++)
				if (*cp < ' ')
					 *cp = ' ';
		    *cp = '\n';
		    *(cp+1) = '\0';

			fbp = fep_buf;
			updateScreen(VALID);
		}

		*k1 = *fbp++;
		if (IsKanji(*k1))
			*k2 = *fbp++;
		if (!*fbp)
			KanjiInput = FALSE;
		return;
	}
#endif /* ONEW */
}

/*
 * get one kanji character from terminal
 */
	void
fep_getc1(cp1, cp2)
	char *cp1, *cp2;
{
	KanjiInput = TRUE;
	fep_mode_switch(-1);
	fep_getchar(cp1, cp2);
#ifdef ONEW
	updateline();
	screen_msg(0, NULL);
	smsg("");
	hira_buf[0] =
#endif ONEW
	*(fbp = fep_buf)  = NUL;
	Kconvlnum     = 0;
	KanjiInput = FALSE;
	showmode();
}

#endif /* FEP */


